from typing import Any, List, Literal

from .....utils.class_registry import register_class

from ....action import CreateDiceAction

from ....event import SkillEndEventArguments
from ....consts import (
    ELEMENT_TO_DIE_COLOR,
    ElementType,
    ObjectPositionType,
    SkillType,
    WeaponType,
)
from ....struct import Cost
from .base import RoundEffectWeaponBase


class SacrificialWeapons_3_3(RoundEffectWeaponBase):
    name: Literal[
        "Sacrificial Fragments",
        "Sacrificial Greatsword",
        "Sacrificial Sword",
        "Sacrificial Bow",
    ]
    cost: Cost = Cost(same_dice_number=3)
    version: Literal["3.3"] = "3.3"
    weapon_type: WeaponType = WeaponType.OTHER
    max_usage_per_round: int = 1

    def __init__(self, *argv, **kwargs):
        super().__init__(*argv, **kwargs)
        if self.name == "Sacrificial Fragments":
            self.weapon_type = WeaponType.CATALYST
        elif self.name == "Sacrificial Bow":
            self.weapon_type = WeaponType.BOW
        elif self.name == "Sacrificial Sword":
            self.weapon_type = WeaponType.SWORD
        else:
            assert self.name == "Sacrificial Greatsword"
            self.weapon_type = WeaponType.CLAYMORE

    def event_handler_SKILL_END(
        self, event: SkillEndEventArguments, match: Any
    ) -> List[CreateDiceAction]:
        """
        If self use elemental skill, create one corresponding dice
        """
        if not (
            self.position.area == ObjectPositionType.CHARACTER
            and event.action.position.player_idx == self.position.player_idx
            and event.action.position.character_idx == self.position.character_idx
            and event.action.skill_type == SkillType.ELEMENTAL_SKILL
            and self.usage > 0
        ):
            # not equipped or not self use elemental skill or no usage
            return []
        self.usage -= 1
        character = match.player_tables[self.position.player_idx].characters[
            self.position.character_idx
        ]
        ele_type: ElementType = character.element
        die_color = ELEMENT_TO_DIE_COLOR[ele_type]
        return [
            CreateDiceAction(
                player_idx=self.position.player_idx,
                number=1,
                color=die_color,
            )
        ]


register_class(SacrificialWeapons_3_3)



from typing import Any, List, Literal

from ....utils.class_registry import register_class

from ...struct import Cost

from ...modifiable_values import DamageDecreaseValue, DamageIncreaseValue, DamageValue

from ...consts import (
    DamageElementalType,
    DamageType,
    IconType,
    ObjectPositionType,
    SkillType,
)

from ...action import MakeDamageAction, RemoveObjectAction

from ...event import MakeDamageEventArguments, SkillEndEventArguments
from .base import (
    DefendTeamStatus,
    ExtraAttackTeamStatus,
    RoundTeamStatus,
    UsageTeamStatus,
)


class SparksNSplash_3_4(UsageTeamStatus):
    name: Literal["Sparks 'n' Splash"] = "Sparks 'n' Splash"
    version: Literal["3.4"] = "3.4"
    usage: int = 2
    max_usage: int = 2
    icon_type: Literal[IconType.OTHERS] = IconType.OTHERS

    def event_handler_SKILL_END(
        self, event: SkillEndEventArguments, match: Any
    ) -> List[MakeDamageAction]:
        """
        When attached character use skill, then damage itself.
        """
        if not self.position.check_position_valid(
            event.action.position,
            match,
            player_idx_same=True,
            target_area=ObjectPositionType.SKILL,
        ):
            # not character use skill, not modify
            return []
        active_character = match.player_tables[
            self.position.player_idx
        ].get_active_character()
        if self.usage > 0:  # pragma: no branch
            self.usage -= 1
            return [
                MakeDamageAction(
                    damage_value_list=[
                        DamageValue(
                            position=self.position,
                            damage_type=DamageType.DAMAGE,
                            target_position=active_character.position,
                            damage=2,
                            damage_elemental_type=DamageElementalType.PYRO,
                            cost=Cost(),
                        )
                    ]
                )
            ]
        else:
            return []  # pragma: no cover


class InspirationField_3_3(RoundTeamStatus):
    name: Literal["Inspiration Field"] = "Inspiration Field"
    desc: Literal["", "talent"] = ""
    version: Literal["3.3"] = "3.3"
    usage: int = 2
    max_usage: int = 2
    icon_type: Literal[IconType.OTHERS] = IconType.OTHERS

    talent_activated: bool = False

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        if self.talent_activated:
            self.desc = "talent"

    def renew(self, new_status: "InspirationField_3_3") -> None:
        super().renew(new_status)
        if new_status.talent_activated:
            self.talent_activated = True
            self.desc = "talent"

    def value_modifier_DAMAGE_INCREASE(
        self, value: DamageIncreaseValue, match: Any, mode: Literal["TEST", "REAL"]
    ) -> DamageIncreaseValue:
        """
        If character HP greater than 7, or talent activated, increase damage
        by 2
        """
        assert mode == "REAL"
        if not value.is_corresponding_character_use_damage_skill(
            self.position, match, None
        ):
            # not our character use skill, do nothing
            return value
        character = match.player_tables[value.position.player_idx].characters[
            value.position.character_idx
        ]
        if not self.talent_activated and character.hp < 7:
            # no talent and hp less than 7, do nothing
            return value
        # increase damage by 2
        value.damage += 2
        return value

    def event_handler_SKILL_END(
        self, event: SkillEndEventArguments, match: Any
    ) -> List[MakeDamageAction]:
        """
        If our character use skill and HP less than 7, heal 2 HP
        """
        if self.position.player_idx != event.action.position.player_idx:
            # not our character, do nothing
            return []
        character = match.player_tables[event.action.position.player_idx].characters[
            event.action.position.character_idx
        ]
        if character.hp > 6:
            # HP greater than 6, do nothing
            return []
        # heal this character by 2
        return [
            MakeDamageAction(
                damage_value_list=[
                    DamageValue(
                        position=self.position,
                        damage_type=DamageType.HEAL,
                        target_position=character.position,
                        damage=-2,
                        damage_elemental_type=DamageElementalType.HEAL,
                        cost=Cost(),
                    )
                ],
            )
        ]


class AurousBlaze_3_3(RoundTeamStatus, ExtraAttackTeamStatus):
    name: Literal["Aurous Blaze"] = "Aurous Blaze"
    version: Literal["3.3"] = "3.3"
    usage: int = 2
    max_usage: int = 2
    icon_type: Literal[IconType.OTHERS] = IconType.OTHERS

    trigger_skill_type: SkillType | None = None
    damage: int = 1
    damage_elemental_type: DamageElementalType = DamageElementalType.PYRO
    decrease_usage: bool = False

    def event_handler_SKILL_END(
        self, event: SkillEndEventArguments, match: Any
    ) -> List[MakeDamageAction]:
        """
        if skill used by self player but not yoimiya, make damage to opponent
        active character.
        """
        character = match.player_tables[self.position.player_idx].get_active_character()
        if character.name == "Yoimiya":
            # yoimiya use skill
            return []
        return super().event_handler_SKILL_END(event, match)


class Pyronado_3_3(UsageTeamStatus, ExtraAttackTeamStatus):
    name: Literal["Pyronado"] = "Pyronado"
    version: Literal["3.3"] = "3.3"
    usage: int = 2
    max_usage: int = 2
    icon_type: Literal[IconType.OTHERS] = IconType.OTHERS

    trigger_skill_type: SkillType | None = None
    damage: int = 2
    damage_elemental_type: DamageElementalType = DamageElementalType.PYRO
    decrease_usage: bool = True

    def event_handler_SKILL_END(
        self, event: SkillEndEventArguments, match: Any
    ) -> List[MakeDamageAction]:
        skill = match.get_object(event.action.position)
        if skill.name == "Pyronado":
            # Xiangling use burst Pyronado
            return []
        return super().event_handler_SKILL_END(event, match)


class FierySanctumField_4_1(DefendTeamStatus):
    name: Literal["Fiery Sanctum Field"] = "Fiery Sanctum Field"
    version: Literal["4.1"] = "4.1"
    usage: int = 1
    max_usage: int = 1
    min_damage_to_trigger: int = 1
    max_in_one_time: int = 1
    decrease_usage_by_damage: bool = False

    def _find_dehya(self, match: Any) -> int:
        """
        Find first alive and standby dehya. If not found, return -1.
        """
        characters = match.player_tables[self.position.player_idx].characters
        active_idx = match.player_tables[self.position.player_idx].active_character_idx
        for cidx, character in enumerate(characters):
            if character.name == "Dehya" and character.is_alive and cidx != active_idx:
                return cidx
        return -1

    def value_modifier_DAMAGE_DECREASE(
        self, value: DamageDecreaseValue, match: Any, mode: Literal["TEST", "REAL"]
    ) -> DamageDecreaseValue:
        """
        Check if have alive dehya on standby, if not, no effect.
        """
        dehya_idx = self._find_dehya(match)
        if dehya_idx == -1:
            # not found dehya
            return value
        return super().value_modifier_DAMAGE_DECREASE(value, match, mode)

    def event_handler_MAKE_DAMAGE(
        self, event: MakeDamageEventArguments, match: Any
    ) -> List[MakeDamageAction | RemoveObjectAction]:
        """
        If self.shield_triggered, then check whether need to deal 1 piercing
        damage to dehya.
        """
        ret: List[MakeDamageAction | RemoveObjectAction] = []
        if self.usage == 0:
            # check if should attack dehya
            dehya_idx = self._find_dehya(match)
            assert dehya_idx != -1
            dehya = match.player_tables[self.position.player_idx].characters[dehya_idx]
            if dehya.hp >= 7 and dehya.is_alive:
                # after shield triggered, dehya has at least 7 hp and alive,
                # make 1 piercing damage to dehya
                ret.append(
                    MakeDamageAction(
                        damage_value_list=[
                            DamageValue(
                                position=self.position,
                                target_position=dehya.position,
                                damage=1,
                                damage_type=DamageType.DAMAGE,
                                damage_elemental_type=DamageElementalType.PIERCING,
                                cost=Cost(),
                            )
                        ]
                    )
                )
        return ret + super().event_handler_MAKE_DAMAGE(event, match)


register_class(
    SparksNSplash_3_4
    | InspirationField_3_3
    | AurousBlaze_3_3
    | Pyronado_3_3
    | FierySanctumField_4_1
)



from lpsim.agents.interaction_agent import InteractionAgent
from lpsim.agents.nothing_agent import NothingAgent
from lpsim.server.match import Match, MatchState
from lpsim.server.deck import Deck
from tests.utils_for_test import (
    check_hp,
    get_test_id_from_command,
    make_respond,
    get_random_state,
    set_16_omni,
)


def test_mona():
    """
    first: quick switch two times and slow switch one time. equip artifact
    and check card can cost less. multiple summon will renew old one.
    Double damage will affect after elemental reaction, and only affect once.
    """
    agent_0 = NothingAgent(player_idx=0)
    agent_1 = InteractionAgent(
        player_idx=1,
        verbose_level=0,
        commands=[
            "sw_card",
            "choose 0",
            "reroll",
            "sw_char 1 2",
            "sw_char 0 6",
            "TEST 7 player 0 not ended",
            "sw_char 2 5",
            "TEST 6 player 0 ended",
            "sw_char 0 4",
            "card 0 0 1 2",
            "skill 1 hydro omni",
            "TEST 4 one summon usage 1",
            "TEST 5 hp 9 10 10",
            "end",
            "reroll 1 2 3 4 5 6 7",
            "skill 1 1 2",
            "skill 1 1 0 2",
            "tune 0 0",
            "tune 1 1",
            "tune 1 2",
            "skill 1 0 1 2",
            "TEST 4 one summon usage 1",
            "end",
            "reroll 1 2 3 4 5 6 7",
            "tune 2 6",
            "tune 3 6",
            "TEST 3 talent can decrease cost",
            "sw_char 1 6",
            "TEST 2 talent cannot decrease cost",
            "sw_char 0 6",
            "TEST 3 talent can decrease cost",
            "end",
            "reroll 1 2 3 4 5 6 7",
            "end",
            "reroll 4 5 6 7",
            "skill 1 2 3",
            "card 0 0 0 1 2",
            "end",
            "TEST 1 hp 0 9 10",
            "reroll 1 3 4 5 6 7",
            "sw_char 2 7",
            "tune 0 1",
            "card 2 0 0 1 2",
            "tune 0 0",
            "tune 2 2",
            "tune 2 2",
            "skill 1 0 1 2",
            "end",
        ],
        only_use_command=True,
    )
    match = Match(version="0.0.1", random_state=get_random_state())
    deck = Deck.from_str(
        """
        character:Mona*2
        character:Fischl
        Prophecy of Submersion*10
        Stellar Predator*10
        Wine-Stained Tricorne*10
        """
    )
    match.set_deck([deck, deck])
    match.config.max_same_card_number = 30
    assert match.start()[0]
    match.step()

    while True:
        if match.need_respond(0):
            make_respond(agent_0, match)
        elif match.need_respond(1):
            while True:
                test_id = get_test_id_from_command(agent_1)
                if test_id == 1:
                    check_hp(match, [[0, 9, 10], [10, 10, 10]])
                elif test_id == 2:
                    hands = match.player_tables[1].hands
                    requests = [
                        x
                        for x in match.requests
                        if x.name == "UseCardRequest"
                        and hands[x.card_idx].name == "Prophecy of Submersion"
                    ]
                    assert len(requests) == 0
                    # assert requests[0].cost.elemental_dice_number == 3
                elif test_id == 3:
                    hands = match.player_tables[1].hands
                    requests = [
                        x
                        for x in match.requests
                        if x.name == "UseCardRequest"
                        and hands[x.card_idx].name == "Prophecy of Submersion"
                    ]
                    assert len(requests) > 0
                    assert requests[0].cost.elemental_dice_number == 2
                elif test_id == 4:
                    summons = match.player_tables[1].summons
                    assert len(summons) == 1
                    assert summons[0].usage == 1
                elif test_id == 5:
                    check_hp(match, [[9, 10, 10], [10, 10, 10]])
                elif test_id == 6:
                    assert match.player_tables[0].has_round_ended
                elif test_id == 7:
                    assert not match.player_tables[0].has_round_ended
                else:
                    break
            make_respond(agent_1, match)
        else:
            raise AssertionError("No need respond.")
        if len(agent_1.commands) == 0:
            break

    assert len(agent_1.commands) == 0
    assert match.round_number == 7
    check_hp(match, [[0, 3, 9], [10, 10, 10]])

    assert match.state != MatchState.ERROR


def test_mona_2():
    """
    when shield of Reflection becomes 0, will not disappear; summon twice
    will recover shield; artifact can effect talent card; attack that
    equip talent will trigger talent; Bubble and talent can work
    simultaneously, with damage: (1 + 1electrocharge + 2) * 2 = 8.
    """
    agent_0 = InteractionAgent(
        player_idx=0,
        verbose_level=0,
        commands=[
            "sw_card 1 3 4",
            "choose 0",
            "reroll",
            "tune 0 7",
            "skill 1 0 1 2",
            "sw_char 1 1",
            "sw_char 0 1",
            "sw_char 1 1",
            "end",
            "reroll",
            "end",
            "reroll",
            "sw_char 2 7",
            "end",
            "reroll",
            "end",
            "TEST 1 hp 7 4 1",
            "reroll 1 2 3 4 5 6 7",
            "sw_char 1 7",
            "choose 0",
            "skill 0 0 1 5",
            "skill 1 0 1 2",
            "end",
            "reroll",
            "skill 0 0 6 7",
        ],
        only_use_command=True,
    )
    agent_1 = InteractionAgent(
        player_idx=1,
        verbose_level=0,
        commands=[
            "sw_card",
            "choose 0",
            "reroll 2 3 4 5 6 7",
            "skill 1 1 2 3",
            "card 0 0 3 4",
            "tune 3 2",
            "skill 1 0 1",
            "end",
            "reroll 3 4 5 6 7",
            "skill 0 0 1",
            "sw_char 2 0",
            "tune 1 0",
            "tune 1 2",
            "card 2 0 0 1 2",
            "sw_char 0 0",
            "end",
            "reroll 1 2 3 4 5 6 7",
            "end",
            "reroll 1 2 3 4 5 6 7",
            "card 1 0 0 1",
            "end",
            "reroll 4 5 6 7",
            "skill 1 2 3",
            "tune 0 3",
            "skill 1 0 1 2",
            "end",
            "reroll 2 3 4 5 6 7",
            "sw_char 2 7",
            "tune 2 6",
            "tune 2 6",
            "card 0 2 5 6",
            "skill 1 0 1 2",
        ],
        only_use_command=True,
    )
    match = Match(version="0.0.1", random_state=get_random_state())
    deck = Deck.from_str(
        """
        character:Mona*2
        character:Fischl
        Prophecy of Submersion*10
        Stellar Predator*10
        Wine-Stained Tricorne*10
        """
    )
    match.set_deck([deck, deck])
    match.config.max_same_card_number = 30
    assert match.start()[0]
    match.step()

    while True:
        if match.need_respond(0):
            while True:
                test_id = get_test_id_from_command(agent_0)
                if test_id == 1:
                    check_hp(match, [[7, 4, 1], [9, 10, 10]])
                else:
                    break
            make_respond(agent_0, match)
        elif match.need_respond(1):
            make_respond(agent_1, match)
        else:
            raise AssertionError("No need respond.")
        if len(agent_1.commands) == 0 and len(agent_0.commands) == 0:
            break

    assert len(agent_0.commands) == 0 and len(agent_1.commands) == 0
    assert match.round_number == 6
    check_hp(match, [[3, 0, 0], [8, 10, 9]])

    assert match.state != MatchState.ERROR


def test_mona_q_enemy_attack():
    agent_0 = InteractionAgent(
        player_idx=0,
        verbose_level=0,
        commands=[
            "sw_card",
            "choose 0",
            "skill 1 0 1 2",
            "end",
            "skill 1 0 1 2",
        ],
        only_use_command=True,
    )
    agent_1 = InteractionAgent(
        player_idx=1,
        verbose_level=0,
        commands=[
            "sw_card 0 1 2",
            "choose 0",
            "skill 1 0 1 2",
            "skill 1 0 1 2",
            "skill 1 0 1 2",
            "card 4 0 0 1 2",
            "card 0 0",
            "end",
        ],
        only_use_command=True,
    )
    match = Match(random_state=get_random_state())
    deck = Deck.from_str(
        """
        character:Mona*2
        character:Fischl
        Sweet Madame*15
        Prophecy of Submersion*15
        """
    )
    match.set_deck([deck, deck])
    match.config.max_same_card_number = 30
    set_16_omni(match)
    match.config.random_first_player = False
    assert match.start()[0]
    match.step()

    while True:
        if match.need_respond(0):
            make_respond(agent_0, match)
        elif match.need_respond(1):
            make_respond(agent_1, match)
        else:
            raise AssertionError("No need respond.")
        if len(agent_1.commands) == 0 and len(agent_0.commands) == 0:
            break

    assert len(agent_0.commands) == 0 and len(agent_1.commands) == 0
    assert match.round_number == 2
    assert len(match.player_tables[1].team_status) == 1
    assert match.player_tables[1].team_status[0].name == "Illusory Bubble"
    check_hp(match, [[3, 10, 10], [9, 10, 10]])

    assert match.state != MatchState.ERROR


if __name__ == "__main__":
    test_mona()
    test_mona_q_enemy_attack()



from lpsim.agents.interaction_agent import InteractionAgent
from lpsim.server.match import Match, MatchState
from lpsim.server.deck import Deck
from tests.utils_for_test import (
    check_usage,
    get_random_state,
    get_test_id_from_command,
    make_respond,
    set_16_omni,
)


def test_seelie():
    cmd_records = [
        [
            "sw_card",
            "choose 2",
            "skill 1 0 1 2",
            "skill 1 0 1 2",
            "skill 1 0 1 2",
            "card 0 0 0",
            "skill 1 0 1 2",
            "skill 0 0 1 2",
            "end",
            "skill 0 0 1 2",
            "TEST 2 card number 9 10",
            "end",
        ],
        [
            "sw_card",
            "choose 2",
            "skill 1 0 1 2",
            "skill 1 0 1 2",
            "card 0 0 0",
            "card 0 0 0",
            "skill 1 0 1 2",
            "TEST 1 all support usage 1",
            "skill 1 0 1 2",
            "end",
            "sw_char 0 0",
            "TEST 3 card number 9 5",
            "skill 0 0 1 2",
        ],
    ]
    agent_0 = InteractionAgent(
        player_idx=0, verbose_level=0, commands=cmd_records[0], only_use_command=True
    )
    agent_1 = InteractionAgent(
        player_idx=1, verbose_level=0, commands=cmd_records[1], only_use_command=True
    )
    # initialize match. It is recommended to use default random state to make
    # replay unchanged.
    match = Match(random_state=get_random_state())
    # deck information
    deck = Deck.from_str(
        """
        default_version:4.0
        character:Noelle*3
        Treasure-Seeking Seelie*30
        """
    )
    match.set_deck([deck, deck])
    match.config.max_same_card_number = None
    match.config.character_number = None
    match.config.card_number = None
    match.config.check_deck_restriction = False
    # check whether random_first_player is enabled.
    match.config.random_first_player = False
    # check whether in rich mode (16 omni each round)
    set_16_omni(match)
    match.start()
    match.step()

    while True:
        if match.need_respond(0):
            agent = agent_0
        elif match.need_respond(1):
            agent = agent_1
        else:
            raise AssertionError("No need respond.")
        # do tests
        while True:
            test_id = get_test_id_from_command(agent)
            if test_id == 0:
                # id 0 means current command is not a test command.
                break
            elif test_id == 1:
                num = [1, 2]
                for table, n in zip(match.player_tables, num):
                    assert n == len(table.supports)
                    for support in table.supports:
                        assert support.usage == 1
            elif test_id == 2:
                assert len(match.player_tables[0].hands) == 9
                assert len(match.player_tables[1].hands) == 10
            elif test_id == 3:
                assert len(match.player_tables[0].hands) == 9
                assert len(match.player_tables[1].hands) == 5
            else:
                raise AssertionError(f"Unknown test id {test_id}")
        # respond
        make_respond(agent, match)
        if len(agent_1.commands) == 0 and len(agent_0.commands) == 0:
            break

    # simulate ends, check final state
    assert match.state != MatchState.ERROR


def test_NRE():
    cmd_records = [
        [
            "sw_card 1",
            "choose 0",
            "skill 1 15 14 13 12 11",
            "card 0 0 10 9",
            "card 3 0",
            "sw_char 1 8",
            "end",
            "card 4 0",
            "TEST 1 p0 support usage 0",
            "TEST 2 p0 card 7",
            "end",
        ],
        [
            "sw_card 1 3",
            "choose 0",
            "sw_char 1 15",
            "sw_char 2 14",
            "card 1 2",
            "TEST 2 p1 hand 4",
            "card 0 0 13 12",
            "TEST 2 p1 card 4",
            "card 0 0 11 10",
            "TEST 2 p1 card 4",
            "card 0 0 9 8",
            "TEST 2 p1 card 4",
            "TEST 1 p1 support usage 0 0",
            "skill 1 7 6 5",
            "TEST 2 p0 card 5",
            "end",
            "card 3 0 15 14",
            "TEST 2 p0 card 7",
            "TEST 2 p1 card 6",
            "TEST 1 p1 support usage 0 0",
            "TEST 1 p0 support usage 1",
            "end",
            "card 5 0 15 14",
            "card 4 0 13 12",
            "TEST 2 p1 card 6",
            "end",
        ],
    ]
    agent_0 = InteractionAgent(
        player_idx=0, verbose_level=0, commands=cmd_records[0], only_use_command=True
    )
    agent_1 = InteractionAgent(
        player_idx=1, verbose_level=0, commands=cmd_records[1], only_use_command=True
    )
    # initialize match. It is recommended to use default random state to make
    # replay unchanged.
    match = Match(random_state=get_random_state())
    # deck information
    deck = Deck.from_str(
        """
        default_version:4.0
        character:Electro Hypostasis
        character:Klee
        character:Keqing
        NRE*15
        Sweet Madame*2
        Mushroom Pizza*2
        Tandoori Roast Chicken*2
        """
    )
    match.set_deck([deck, deck])
    match.config.max_same_card_number = None
    match.config.character_number = None
    match.config.card_number = None
    match.config.check_deck_restriction = False
    # check whether random_first_player is enabled.
    match.config.random_first_player = False
    # check whether in rich mode (16 omni each round)
    set_16_omni(match)
    match.start()
    match.step()

    while True:
        if match.need_respond(0):
            agent = agent_0
        elif match.need_respond(1):
            agent = agent_1
        else:
            raise AssertionError("No need respond.")
        # do tests
        while True:
            cmd = agent.commands[0]
            test_id = get_test_id_from_command(agent)
            if test_id == 0:
                # id 0 means current command is not a test command.
                break
            elif test_id == 1:
                cmd = cmd.split()
                pidx = int(cmd[2][1])
                usage = [int(x) for x in cmd[5:]]
                supports = match.player_tables[pidx].supports
                assert len(usage) == len(supports)
                for u, s in zip(usage, supports):
                    assert u == s.usage
            elif test_id == 2:
                cmd = cmd.split()
                pidx = int(cmd[2][1])
                cardnum = int(cmd[4])
                assert len(match.player_tables[pidx].hands) == cardnum
            else:
                raise AssertionError(f"Unknown test id {test_id}")
        # respond
        make_respond(agent, match)
        if len(agent_1.commands) == 0 and len(agent_0.commands) == 0:
            break

    # simulate ends, check final state
    assert match.state != MatchState.ERROR


def test_parametric():
    cmd_records = [
        [
            "sw_card",
            "choose 0",
            "card 0 0 15 14",
            "skill 1 13 12 11",
            "TEST 1 p0 support 1",
            "TEST 1 p1 support 0",
            "sw_char 2 10",
            "skill 0 9 8 7",
            "skill 2 6 5 4 3 2",
            "end",
            "skill 0 15 14 13",
            "TEST 2 p1 dice 13",
            "card 0 0 12 11",
            "card 0 0 10 9",
            "card 0 0 8 7",
            "skill 1 6 5 4",
            "skill 1 3 2 1",
            "end",
            "end",
        ],
        [
            "sw_card",
            "choose 1",
            "TEST 1 p0 support 1",
            "card 2 0 15 14",
            "skill 1 13 12 11",
            "TEST 1 p0 support 1",
            "TEST 1 p1 support 0",
            "sw_char 0 10",
            "TEST 1 p0 support 2",
            "TEST 1 p1 support 1",
            "end",
            "TEST 1 p0 support 2",
            "TEST 1 p1 support 1",
            "skill 0 15 14 13",
            "skill 1 12 11 10",
            "skill 1 9 8 7",
            "TEST 2 p0 dice 10",
            "end",
            "sw_char 3 0",
            "card 0 0 0 1",
            "skill 1 0 1 2",
            "end",
        ],
    ]
    agent_0 = InteractionAgent(
        player_idx=0, verbose_level=0, commands=cmd_records[0], only_use_command=True
    )
    agent_1 = InteractionAgent(
        player_idx=1, verbose_level=0, commands=cmd_records[1], only_use_command=True
    )
    # initialize match. It is recommended to use default random state to make
    # replay unchanged.
    match = Match(random_state=get_random_state())
    # deck information
    deck = Deck.from_str(
        """
        default_version:4.0
        character:Mona
        character:Xiangling
        character:Ganyu
        character:Barbara
        Parametric Transformer*30
        """
    )
    match.set_deck([deck, deck])
    match.config.max_same_card_number = None
    match.config.character_number = None
    match.config.card_number = None
    match.config.check_deck_restriction = False
    # check whether random_first_player is enabled.
    match.config.random_first_player = False
    # check whether in rich mode (16 omni each round)
    set_16_omni(match)
    match.start()
    match.step()

    while True:
        if match.need_respond(0):
            agent = agent_0
        elif match.need_respond(1):
            agent = agent_1
        else:
            raise AssertionError("No need respond.")
        # do tests
        while True:
            cmd = agent.commands[0]
            test_id = get_test_id_from_command(agent)
            if test_id == 0:
                # id 0 means current command is not a test command.
                break
            elif test_id == 1:
                cmd = cmd.split()
                pidx = int(cmd[2][1])
                usages = [int(x) for x in cmd[4:]]
                supports = match.player_tables[pidx].supports
                assert len(supports) == len(usages)
                for u, s in zip(usages, supports):
                    assert u == s.usage
            elif test_id == 2:
                cmd = cmd.split()
                pidx = int(cmd[2][1])
                num = int(cmd[-1])
                assert len(match.player_tables[pidx].dice.colors) == num
            else:
                raise AssertionError(f"Unknown test id {test_id}")
        # respond
        make_respond(agent, match)
        if len(agent_1.commands) == 0 and len(agent_0.commands) == 0:
            break

    # simulate ends, check final state
    assert match.state != MatchState.ERROR


def test_red_feather_fan():
    cmd_records = [
        [
            "sw_card 1 3",
            "choose 0",
            "sw_char 1 15",
            "TEST 1 p1 support 1 1",
            "card 1 0 14 13",
            "card 0 0",
            "card 1 0",
            "sw_char 2",
            "sw_char 1",
            "sw_char 0 12",
            "TEST 1 p1 support 0 0",
            "TEST 2 p1 dice 10",
            "end",
            "sw_char 1 15",
            "skill 1 14 13 12",
            "TEST 1 p1 support 0 0",
            "TEST 2 p1 dice 14",
            "card 1 0",
            "end",
            "sw_char 2",
            "card 1 0",
            "card 2 0",
            "sw_char 1",
            "TEST 1 p0 support 1",
            "sw_char 0",
        ],
        [
            "sw_card 2 3",
            "choose 0",
            "card 0 0 15 14",
            "card 0 0 13 12",
            "TEST 3 sw cost 1",
            "sw_char 1 11",
            "TEST 2 p0 dice 12",
            "TEST 1 p0 support 0",
            "sw_char 0",
            "card 0 0",
            "sw_char 2",
            "sw_char 1 10",
            "end",
            "sw_char 2 15",
            "card 1 0",
            "sw_char 1",
            "TEST 1 p1 support 0 1",
            "card 1 0",
            "sw_char 0",
            "sw_char 1 14",
            "end",
            "sw_char 2 15",
        ],
    ]
    agent_0 = InteractionAgent(
        player_idx=0, verbose_level=0, commands=cmd_records[0], only_use_command=True
    )
    agent_1 = InteractionAgent(
        player_idx=1, verbose_level=0, commands=cmd_records[1], only_use_command=True
    )
    # initialize match. It is recommended to use default random state to make
    # replay unchanged.
    match = Match(random_state=get_random_state())
    # deck information
    deck = Deck.from_str(
        """
        default_version:4.0
        character:Diona
        character:Noelle
        character:Sucrose
        Red Feather Fan*10
        Changing Shifts*10
        Leave It to Me!*10
        """
    )
    match.set_deck([deck, deck])
    match.config.max_same_card_number = None
    match.config.character_number = None
    match.config.card_number = None
    match.config.check_deck_restriction = False
    # check whether random_first_player is enabled.
    match.config.random_first_player = False
    # check whether in rich mode (16 omni each round)
    set_16_omni(match)
    match.start()
    match.step()

    while True:
        if match.need_respond(0):
            agent = agent_0
        elif match.need_respond(1):
            agent = agent_1
        else:
            raise AssertionError("No need respond.")
        # do tests
        while True:
            cmd = agent.commands[0]
            test_id = get_test_id_from_command(agent)
            if test_id == 0:
                # id 0 means current command is not a test command.
                break
            elif test_id == 1:
                # a sample of HP check based on the command string.
                cmd = cmd.split()
                pidx = int(cmd[2][1])
                check_usage(match.player_tables[pidx].supports, cmd[4:])
            elif test_id == 2:
                cmd = cmd.split()
                pidx = int(cmd[2][1])
                assert len(match.player_tables[pidx].dice.colors) == int(cmd[-1])
            elif test_id == 3:
                for req in match.requests:
                    if req.name == "SwitchCharacterRequest":
                        assert req.cost.total_dice_cost == 1
            else:
                raise AssertionError(f"Unknown test id {test_id}")
        # respond
        make_respond(agent, match)
        if len(agent_1.commands) == 0 and len(agent_0.commands) == 0:
            break

    # simulate ends, check final state
    assert match.state != MatchState.ERROR


if __name__ == "__main__":
    test_seelie()
    test_NRE()
    test_parametric()
    test_red_feather_fan()



"""
Collect descriptions.
"""
import json
import logging
from typing import Any, Dict, Literal, TypedDict


ExpectedLanguageType = Literal["zh-CN", "en-US"]


class DescDictType(TypedDict, total=False):
    """
    image_path: optional, image path
    id: optional, id for sorting cards
    names: name of the card, key is language type, value is name
    descs: description of the card, key is version, value is dict of language
        type and description
    cost: cost of the card, key is version, value is dict of Cost type
    """

    image_path: str
    id: int
    names: Dict[ExpectedLanguageType, str]
    descs: Dict[str, Dict[ExpectedLanguageType, str]]
    cost: Dict[str, Dict[str, Any]]


_default_json_path = __file__[: -len("desc_registry.py")] + "default_desc.json"
_desc_dict: Dict[str, DescDictType] = json.load(
    open(_default_json_path, "r", encoding="utf-8")
)


def _parse_value(value: Any):
    """
    If value is str and starts with $: it is a reference to current desc,
    return its referenced result. Reference key format is:
    $key1|key2|key3 ...
    """
    if isinstance(value, str) and value.startswith("$"):
        # return referenced result
        keys = value[1:].split("|")
        root = _desc_dict
        for key in keys:
            if key not in root:
                raise ValueError(f'in "{value}": key "{key}" not found')
            root = root[key]
        return root
    return value


def _merge_dict(source, destination, prev_key: str = ""):
    for key, value in source.items():
        if prev_key == "":
            pkey = key
        else:
            pkey = f"{prev_key}|{key}"
        if isinstance(value, dict):
            # get node or create one
            node = destination.setdefault(key, {})
            _merge_dict(value, node, prev_key=pkey)
        else:
            value = _parse_value(value)
            if "|names|" in pkey and value in ["", "XXX"]:
                # name is empty, raise error.
                raise ValueError(f'name value "{value}" of key "{pkey}" is empty')
            if key in destination:
                if destination[key] != value:
                    raise ValueError(
                        f'value "{destination[key]}" of key "{pkey}" '
                        f'is replaced by new value "{value}"'
                    )
                else:
                    logging.warning(f'duplicate adding key "{pkey}"')
            destination[key] = value
    return destination


def update_desc(desc_dict: Dict[str, DescDictType]) -> None:
    """
    Update descriptions.
    """
    for key in desc_dict.keys():
        if (
            key[:16] == "CHARACTER_STATUS"
            or key[:5] == "SKILL"
            or key[:6] == "SUMMON"
            or key[:11] == "TEAM_STATUS"
        ):
            if "id" in desc_dict[key]:
                raise ValueError(f"id should not be in {key}")
    _merge_dict(desc_dict, _desc_dict)


def desc_exist(
    type: str,
    name: str,
    version: str,
    check_id: bool | None = None,
    check_image_path: bool | None = None,
) -> bool:
    """
    Check if the description of the class exists.
    Also check if id and image_path exists if check_id and check_image_path is not None,
    when not satisfied for these two, a warning message will be shown, but will not
    change the return.
    """
    full_name = f"{type}/{name}"
    if full_name not in _desc_dict:
        return False
    if check_id is not None:
        if ("id" in _desc_dict[full_name]) != check_id:
            w = "with" if check_id else "without"
            logging.warning(f"expected {full_name} {w} id, but not satisfied")
    if check_image_path is not None:
        if ("image_path" in _desc_dict[full_name]) != check_image_path:
            w = "with" if check_image_path else "without"
            logging.warning(f"expected {full_name} {w} image_path, but not satisfied")
    return (
        full_name in _desc_dict
        and "descs" in _desc_dict[full_name]
        and version in _desc_dict[full_name]["descs"]  # type: ignore
    )


def get_desc_patch() -> Dict[str, DescDictType]:
    """
    Get desc patch.
    """
    return _desc_dict


def update_cost(type: str, name: str, version: str, cost: Any) -> None:
    """
    update cost into desc dict.
    """
    if not desc_exist(type, name, version):
        raise ValueError(f"cannot find desc for {type}/{name}/{version}")
    full_name = f"{type}/{name}"
    d = _desc_dict[full_name]
    if "cost" not in _desc_dict[full_name]:
        d["cost"] = {}
    d["cost"][version] = cost.dict()  # type: ignore


__all__ = [
    "DescDictType",
    "ExpectedLanguageType",
    "update_desc",
    "desc_exist",
    "get_desc_patch",
    "update_cost",
]



"""
Event handlers to implement system controls and special rules 
"""
from typing import Dict, List, Literal, Any

from ..utils.desc_registry import DescDictType

from ..utils.class_registry import register_base_class, register_class
from .object_base import ObjectBase
from .struct import ObjectPosition
from .consts import ObjectPositionType, DieColor, ObjectType
from .event import (
    AfterMakeDamageEventArguments,
    CharacterDefeatedEventArguments,
    RoundEndEventArguments,
    DrawCardEventArguments,
    RoundPrepareEventArguments,
    SkillEndEventArguments,
)
from .action import (
    Actions,
    CharacterDefeatedAction,
    GenerateChooseCharacterRequestAction,
    DrawCardAction,
    RemoveCardAction,
)
from .modifiable_values import InitialDiceColorValue, RerollValue


class SystemEventHandlerBase(ObjectBase):
    """
    Base class of system event handlers. Its position should be SYSTEM.
    """

    name: Literal["SystemEventHandlerBase"]
    type: Literal[ObjectType.SYSTEM] = ObjectType.SYSTEM

    position: ObjectPosition = ObjectPosition(
        player_idx=-1,
        area=ObjectPositionType.SYSTEM,
        id=-1,
    )

    def renew(self, target: "SystemEventHandlerBase") -> None:  # pragma: no cover
        """
        No need to renew.
        """
        pass


register_base_class(SystemEventHandlerBase)


class SystemEventHandler_3_3(SystemEventHandlerBase):
    name: Literal["System"] = "System"
    version: Literal["3.3"] = "3.3"

    def event_handler_DRAW_CARD(
        self, event: DrawCardEventArguments, match: Any
    ) -> List[Actions]:
        """
        After draw card, check whether max card number is exceeded.
        """
        player_idx = event.action.player_idx
        hand_size = event.hand_size
        max_hand_size = event.max_hand_size
        actions = []
        hands = match.player_tables[player_idx].hands
        if hand_size > max_hand_size:
            for i in range(hand_size, max_hand_size, -1):
                # remove the last card in hand repeatedly until hand size
                # is less than max hand size
                actions.append(
                    RemoveCardAction(
                        position=hands[i - 1].position, remove_type="BURNED"
                    )
                )
        return actions

    def event_handler_AFTER_MAKE_DAMAGE(
        self, event: AfterMakeDamageEventArguments, match: Any
    ) -> List[Actions]:
        """
        After damage made, check whether anyone has defeated.
        """
        actions: List[Actions] = []
        for pnum, table in enumerate(match.player_tables):
            for cnum, character in enumerate(table.characters):
                if character.hp == 0 and character.is_alive:
                    character.is_alive = False  # mark as defeated first
                    actions.append(
                        CharacterDefeatedAction(
                            player_idx=pnum,
                            character_idx=cnum,
                        )
                    )
        return actions

    def event_handler_CHARACTER_DEFEATED(
        self, event: CharacterDefeatedEventArguments, match: Any
    ) -> List[Actions]:
        """
        After character defeated, check whether need to switch character.
        """
        if event.need_switch:
            return [
                GenerateChooseCharacterRequestAction(
                    player_idx=event.action.player_idx,
                )
            ]
        return []

    def event_handler_ROUND_END(
        self, event: RoundEndEventArguments, match: Any
    ) -> List[Actions]:
        """
        After round end, draw card for each player.
        """
        actions: List[Actions] = []
        first = event.player_go_first
        initial_card_draw = event.initial_card_draw
        actions.append(
            DrawCardAction(
                player_idx=first,
                number=initial_card_draw,
                draw_if_filtered_not_enough=True,
            )
        )
        actions.append(
            DrawCardAction(
                player_idx=1 - first,
                number=initial_card_draw,
                draw_if_filtered_not_enough=True,
            )
        )
        return actions


class SystemEventHandler_3_4(SystemEventHandler_3_3):
    version: Literal["3.4"] = "3.4"


class OmnipotentGuideEventHandler_3_3(SystemEventHandlerBase):
    name: Literal["Omnipotent Guide"] = "Omnipotent Guide"
    version: Literal["3.3"] = "3.3"

    def value_modifier_INITIAL_DICE_COLOR(
        self, value: InitialDiceColorValue, match: Any, mode: Literal["REAL", "TEST"]
    ) -> InitialDiceColorValue:
        """
        remove current dice color and add 100 omni dice color
        """
        value.dice_colors = [DieColor.OMNI] * 100
        return value

    def value_modifier_REROLL(
        self, value: RerollValue, match: Any, mode: Literal["REAL", "TEST"]
    ) -> RerollValue:
        """
        reroll set to 0
        """
        value.value = 0
        return value


class EllinEventHandler_3_3(SystemEventHandlerBase):
    """
    This handler will record all skill ids used in this round.
    """

    name: Literal["Ellin"] = "Ellin"
    version: Literal["3.3"] = "3.3"
    recorded_skill_ids: Dict[int, None] = {}

    def event_handler_ROUND_PREPARE(
        self, event: RoundPrepareEventArguments, match: Any
    ) -> List[Actions]:
        self.recorded_skill_ids.clear()
        return []

    def event_handler_SKILL_END(
        self, event: SkillEndEventArguments, match: Any
    ) -> List[Actions]:
        """
        Record skill id when skill end.
        """
        self.recorded_skill_ids[event.action.position.id] = None
        return []


class IHaventLostYetEventHandler_3_3(SystemEventHandlerBase):
    """
    This handler will record whether anyone is defeated this round for each
    player.
    """

    name: Literal["I Haven't Lost Yet!"] = "I Haven't Lost Yet!"
    version: Literal["3.3"] = "3.3"
    defeated: List[bool] = [False, False]

    def event_handler_ROUND_PREPARE(
        self, event: RoundPrepareEventArguments, match: Any
    ) -> List[Actions]:
        self.defeated = [False, False]
        return []

    def event_handler_CHARACTER_DEFEATED(
        self, event: CharacterDefeatedEventArguments, match: Any
    ) -> List[Actions]:
        self.defeated[event.action.player_idx] = True
        return []


SystemEventHandler = SystemEventHandler_3_4


event_handler_descs: Dict[str, DescDictType] = {
    "SYSTEM/System": {
        "names": {"zh-CN": "系统事件", "en-US": "System events"},
        "descs": {"3.3": {}, "3.4": {}},
    },
    "SYSTEM/Omnipotent Guide": {
        "names": {"zh-CN": "万能向导", "en-US": "Omnipotent Guide"},
        "descs": {"3.3": {}},
    },
    "SYSTEM/Ellin": {
        "names": {"zh-CN": "艾琳", "en-US": "Ellin"},
        "descs": {"3.3": {}},
    },
    "SYSTEM/I Haven't Lost Yet!": {
        "names": {"zh-CN": "本大爷还没有输！", "en-US": "I Haven't Lost Yet!"},
        "descs": {"3.3": {}},
    },
}


register_class(
    SystemEventHandler_3_3
    | OmnipotentGuideEventHandler_3_3
    | SystemEventHandler_3_4
    | EllinEventHandler_3_3
    | IHaventLostYetEventHandler_3_3,
    event_handler_descs,
)



from lpsim.agents.interaction_agent import InteractionAgent
from lpsim.agents.nothing_agent import NothingAgent
from lpsim.server.match import Match, MatchState
from lpsim.server.consts import DamageElementalType
from lpsim.server.deck import Deck
from tests.utils_for_test import (
    check_hp,
    get_test_id_from_command,
    make_respond,
    get_random_state,
    set_16_omni,
)


def test_fischl_mona_nahida():
    """
    3335 + E + talent
    """
    agent_0 = NothingAgent(player_idx=0)
    agent_1 = InteractionAgent(
        player_idx=1,
        verbose_level=0,
        commands=[
            "sw_card",
            "choose 0",
            "card 1 0 0 1",
            "skill 1 0 1 2",
            "sw_char 1 1",
            "skill 1 omni omni hydro",
            "end",
            "TEST 1 hp 4 8 8",
            "skill 1 0 1 2",
            "sw_char 2 0",
            "skill 2 omni omni dendro omni omni",
            "TEST 2 hp 0 7 7 and 1 seed for alive",
            "end",
            "TEST 3 hp 0 1 5",
            "skill 1 0 1 2",
            "card 0 0 0 1 2",
        ],
        only_use_command=True,
    )
    match = Match(random_state=get_random_state())
    deck = Deck.from_str(
        """
        character:Fischl
        character:Mona
        character:Nahida
        Rana*10
        Wine-Stained Tricorne*10
        The Seed of Stored Knowledge*10
        """
    )
    match.set_deck([deck, deck])
    match.config.max_same_card_number = 30
    set_16_omni(match)
    assert match.start()[0]
    match.step()

    while True:
        if match.need_respond(0):
            make_respond(agent_0, match)
        elif match.need_respond(1):
            while True:
                test_id = get_test_id_from_command(agent_1)
                if test_id == 1:
                    check_hp(match, [[4, 8, 8], [10, 10, 10]])
                elif test_id == 2:
                    check_hp(match, [[0, 7, 7], [10, 10, 10]])
                    for character in match.player_tables[0].characters:
                        if character.is_alive:
                            assert len(character.status) == 1
                            assert character.status[0].name == "Seed of Skandha"
                        else:
                            assert len(character.status) == 0
                elif test_id == 3:
                    check_hp(match, [[0, 1, 5], [10, 10, 10]])
                else:
                    break
            make_respond(agent_1, match)
        else:
            raise AssertionError("No need respond.")
        if len(agent_1.commands) == 0:
            break

    assert len(agent_1.commands) == 0
    assert match.round_number == 3
    check_hp(match, [[0, 0, 1], [10, 10, 10]])
    assert len(match.player_tables[1].team_status) == 1
    assert match.player_tables[1].team_status[0].name == "Shrine of Maya"
    assert match.player_tables[1].team_status[0].usage == 3

    assert match.state != MatchState.ERROR


def test_fischl_mona_nahida_no_talent():
    """
    3335 + EQ
    """
    agent_0 = NothingAgent(player_idx=0)
    agent_1 = InteractionAgent(
        player_idx=1,
        verbose_level=0,
        commands=[
            "sw_card",
            "choose 0",
            "card 1 0 0 1",
            "skill 1 0 1 2",
            "sw_char 1 1",
            "skill 1 omni omni hydro",
            "end",
            "TEST 1 hp 4 8 8",
            "skill 1 0 1 2",
            "sw_char 2 0",
            "skill 2 omni omni dendro omni omni",
            "TEST 2 hp 0 7 7 and 1 seed for alive",
            "end",
            "TEST 3 hp 0 1 5",
            "skill 1 0 1 2",
            "skill 3 0 1 2",
        ],
        only_use_command=True,
    )
    match = Match(random_state=get_random_state())
    deck = Deck.from_str(
        """
        character:Fischl
        character:Mona
        character:Nahida
        Rana*10
        Wine-Stained Tricorne*10
        The Seed of Stored Knowledge*10
        """
    )
    match.set_deck([deck, deck])
    match.config.max_same_card_number = 30
    set_16_omni(match)
    assert match.start()[0]
    match.step()

    while True:
        if match.need_respond(0):
            make_respond(agent_0, match)
        elif match.need_respond(1):
            while True:
                test_id = get_test_id_from_command(agent_1)
                if test_id == 1:
                    check_hp(match, [[4, 8, 8], [10, 10, 10]])
                elif test_id == 2:
                    check_hp(match, [[0, 7, 7], [10, 10, 10]])
                    for character in match.player_tables[0].characters:
                        if character.is_alive:
                            assert len(character.status) == 1
                            assert character.status[0].name == "Seed of Skandha"
                        else:
                            assert len(character.status) == 0
                elif test_id == 3:
                    check_hp(match, [[0, 1, 5], [10, 10, 10]])
                else:
                    break
            make_respond(agent_1, match)
        else:
            raise AssertionError("No need respond.")
        if len(agent_1.commands) == 0:
            break

    assert len(agent_1.commands) == 0
    assert match.round_number == 3
    check_hp(match, [[0, 0, 1], [10, 10, 10]])
    # makes small tests on getting id
    assert match.player_tables[0].previous_character_idx() is None
    assert match.player_tables[0].next_character_idx() is None
    assert match.player_tables[1].previous_character_idx() == 1
    assert match.player_tables[1].next_character_idx() == 0
    for i in range(3):
        assert match.player_tables[1].previous_character_idx(i) == (i + 2) % 3
        assert match.player_tables[1].next_character_idx(i) == (i + 1) % 3
    assert len(match.player_tables[1].team_status) == 1
    assert match.player_tables[1].team_status[0].name == "Shrine of Maya"
    assert match.player_tables[1].team_status[0].usage == 2

    assert match.state != MatchState.ERROR

    assert match.need_respond(1)
    agent_1.commands = ["skill 1 0 1 2"]
    make_respond(agent_1, match)
    assert match.state == MatchState.ENDED
    assert not match.step()


def test_nahida_talents():
    """
    first: with no pyro hydro electro ally
    """
    agent_0 = InteractionAgent(
        player_idx=0,
        verbose_level=0,
        commands=[
            "sw_card",
            "choose 0",
            "sw_char 1 0",
            "sw_char 0 0",
            "end",
            "sw_char 2 0",
            "sw_char 0 0",
            "sw_char 1 0",
            "sw_char 2 0",
            "sw_char 1 0",
            "sw_char 2 0",
            "skill 1 0 1 2",
            "end",
            "end",
        ],
        only_use_command=True,
    )
    agent_1 = InteractionAgent(
        player_idx=1,
        verbose_level=0,
        commands=[
            "sw_card",
            "choose 2",
            "skill 1 0 1 2",
            "TEST 1 only p0c1 2seed",
            "skill 1 0 1 2",
            "TEST 2 p0c0 p0c1 2seed",
            "skill 1 0 1 2",
            "TEST 3 all 2seed hp 6 8 10",
            "end",
            "card 5 0 0 1 2",
            "TEST 4 maya 2 seed all 2",
            "sw_char 0 0",
            "skill 0 0 1 2",
            "TEST 5 hp 5 4 5 mid no element",
            "sw_char 2 0",
            "skill 1 0 1 2",
            "TEST 6 hp 5 2 5 all dendro 2seed",
            "sw_char 1 0",
            "sw_char 2 0",
            "sw_char 1 0",
            "sw_char 2 0",
            "end",
            "TEST 7 hp 4 1 1, ele DDN, all 1seed, burining flame, 1 maya",
            "sw_char 0 0",
        ],
        only_use_command=True,
    )
    match = Match(version="0.0.1", random_state=get_random_state())
    deck1 = Deck.from_str(
        """
        default_version:4.0
        character:ElectroMobMage
        character:PyroMobMage
        character:HydroMobMage
        Rana*10
        Wine-Stained Tricorne*10
        The Seed of Stored Knowledge*10
    """
    )
    deck2 = Deck.from_str(
        """
        character:DendroMobMage
        character:DendroMobMage
        character:Nahida
        Rana*10
        Wine-Stained Tricorne*10
        The Seed of Stored Knowledge*10
        """
    )
    match.set_deck([deck1, deck2])
    match.config.max_same_card_number = 30
    match.config.check_deck_restriction = False
    match.config.random_first_player = False
    set_16_omni(match)
    assert match.start()[0]
    match.player_tables[1].characters[0].skills[
        0
    ].damage_type = DamageElementalType.PYRO  # can do pyro damage
    match.step()

    while True:
        if match.need_respond(0):
            make_respond(agent_0, match)
        elif match.need_respond(1):
            while True:
                test_id = get_test_id_from_command(agent_1)
                if test_id == 1:
                    characters = match.player_tables[0].characters
                    assert len(characters[0].status) == 0
                    assert len(characters[1].status) == 1
                    assert characters[1].status[0].name == "Seed of Skandha"
                    assert characters[1].status[0].usage == 2
                    assert len(characters[2].status) == 0
                elif test_id == 2:
                    characters = match.player_tables[0].characters
                    assert len(characters[0].status) == 1
                    assert characters[0].status[0].name == "Seed of Skandha"
                    assert characters[0].status[0].usage == 2
                    assert len(characters[1].status) == 1
                    assert characters[1].status[0].name == "Seed of Skandha"
                    assert characters[1].status[0].usage == 2
                    assert len(characters[2].status) == 0
                elif test_id == 3:
                    characters = match.player_tables[0].characters
                    for character in characters:
                        assert len(character.status) == 1
                        assert character.status[0].name == "Seed of Skandha"
                        assert character.status[0].usage == 2
                    for character in characters[:2]:
                        assert character.element_application == ["DENDRO"]
                    assert characters[2].element_application == []
                    check_hp(match, [[6, 8, 10], [10, 10, 10]])
                elif test_id == 4:
                    characters = match.player_tables[0].characters
                    for character in characters:
                        assert len(character.status) == 1
                        assert character.status[0].name == "Seed of Skandha"
                        assert character.status[0].usage == 2
                    for character in characters:
                        assert character.element_application == ["DENDRO"]
                    assert len(match.player_tables[1].team_status) == 1
                    assert match.player_tables[1].team_status[0].name == (
                        "Shrine of Maya"
                    )
                    assert match.player_tables[1].team_status[0].usage == 2
                elif test_id == 5:
                    characters = match.player_tables[0].characters
                    for cid, character in enumerate(characters):
                        assert len(character.status) == 1
                        assert character.status[0].name == "Seed of Skandha"
                        assert character.status[0].usage == 1
                        if cid == 1:
                            assert character.element_application == []
                        else:
                            assert character.element_application == ["DENDRO"]
                    check_hp(match, [[5, 4, 5], [10, 10, 10]])
                elif test_id == 6:
                    # TEST 6 hp 5 2 5 all dendro 2seed
                    characters = match.player_tables[0].characters
                    for character in characters:
                        assert len(character.status) == 1
                        assert character.status[0].name == "Seed of Skandha"
                        assert character.status[0].usage == 2
                        assert character.element_application == ["DENDRO"]
                    check_hp(match, [[5, 2, 5], [10, 10, 10]])
                elif test_id == 7:
                    # hp 4 1 1, ele DDN, all 1seed, burining flame, 1 maya
                    characters = match.player_tables[0].characters
                    for cid, character in enumerate(characters):
                        assert len(character.status) == 1
                        assert character.status[0].name == "Seed of Skandha"
                        assert character.status[0].usage == 1
                        if cid == 2:
                            assert character.element_application == []
                        else:
                            assert character.element_application == ["DENDRO"]
                    check_hp(match, [[4, 1, 1], [10, 7, 10]])
                    assert len(match.player_tables[1].team_status) == 1
                    assert match.player_tables[1].team_status[0].name == (
                        "Shrine of Maya"
                    )
                    assert match.player_tables[1].team_status[0].usage == 1
                    assert len(match.player_tables[1].summons) == 1
                    assert match.player_tables[1].summons[0].name == ("Burning Flame")
                    assert match.player_tables[1].summons[0].usage == 1
                else:
                    break
            make_respond(agent_1, match)
        else:
            raise AssertionError("No need respond.")
        if len(agent_1.commands) == 0:
            break

    assert len(agent_1.commands) == 0
    assert match.round_number == 3
    check_hp(match, [[4, 1, 1], [10, 7, 10]])

    assert match.state != MatchState.ERROR


def test_nahida_talents_2():
    """
    second: with pyro electro (hydro tested above)
    """
    agent_0 = InteractionAgent(
        player_idx=0,
        verbose_level=0,
        commands=[
            "sw_card",
            "choose 1",
            "skill 1 0 1 2",
            "sw_char 0 0",
            "end",
            "sw_char 2 0",
            "sw_char 0 0",
            "sw_char 2 0",
            "sw_char 1 0",
            "sw_char 2 0",
            "end",
            "end",
        ],
        only_use_command=True,
    )
    agent_1 = InteractionAgent(
        player_idx=1,
        verbose_level=0,
        commands=[
            "sw_card",
            "choose 2",
            "skill 1 0 1 2",
            "TEST 1 only p0c1 2seed",
            "skill 1 0 1 2",
            "TEST 2 p0c0 p0c1 2seed",
            "skill 1 0 1 2",
            "TEST 3 all 2seed hp 6 8 10",
            "end",
            "card 5 0 0 1 2",
            "TEST 4 maya 2 seed all 3, p1c2 seed 2",
            "skill 1 0 1 2",
            "TEST 4 maya 2 seed all 3, p1c2 seed 2",
            "sw_char 0 0",
            "skill 0 0 1 2",
            "TEST 5 hp 3 4 5 all dendro",
            "sw_char 2 0",
            "end",
            "TEST 7 hp 4 3 1, all dendro, all 1seed, burining flame, 1 maya",
            "sw_char 0 0",
        ],
        only_use_command=True,
    )
    match = Match(version="0.0.1", random_state=get_random_state())
    deck1 = Deck.from_str(
        """
        default_version:4.0
        character:ElectroMobMage
        character:Nahida
        character:HydroMobMage
        Rana*10
        Wine-Stained Tricorne*10
        The Seed of Stored Knowledge*10
    """
    )
    deck2 = Deck.from_str(
        """
        character:PyroMobMage
        character:ElectroMobMage
        character:Nahida
        Rana*10
        Wine-Stained Tricorne*10
        The Seed of Stored Knowledge*10
        """
    )
    match.set_deck([deck1, deck2])
    match.config.max_same_card_number = 30
    match.config.random_first_player = False
    set_16_omni(match)
    assert match.start()[0]
    match.step()

    while True:
        if match.need_respond(0):
            make_respond(agent_0, match)
        elif match.need_respond(1):
            while True:
                test_id = get_test_id_from_command(agent_1)
                if test_id == 1:
                    characters = match.player_tables[0].characters
                    assert len(characters[0].status) == 0
                    assert len(characters[1].status) == 1
                    assert characters[1].status[0].name == "Seed of Skandha"
                    assert characters[1].status[0].usage == 2
                    assert len(characters[2].status) == 0
                elif test_id == 2:
                    characters = match.player_tables[0].characters
                    assert len(characters[0].status) == 1
                    assert characters[0].status[0].name == "Seed of Skandha"
                    assert characters[0].status[0].usage == 2
                    assert len(characters[1].status) == 1
                    assert characters[1].status[0].name == "Seed of Skandha"
                    assert characters[1].status[0].usage == 2
                    assert len(characters[2].status) == 0
                elif test_id == 3:
                    characters = match.player_tables[0].characters
                    for character in characters:
                        assert len(character.status) == 1
                        assert character.status[0].name == "Seed of Skandha"
                        assert character.status[0].usage == 2
                    for character in characters[:2]:
                        assert character.element_application == ["DENDRO"]
                    assert characters[2].element_application == []
                    check_hp(match, [[6, 8, 10], [10, 10, 8]])
                elif test_id == 4:
                    characters = match.player_tables[0].characters
                    for character in characters:
                        assert len(character.status) == 1
                        assert character.status[0].name == "Seed of Skandha"
                        assert character.status[0].usage == 3
                    for character in characters:
                        assert character.element_application == ["DENDRO"]
                    assert len(match.player_tables[1].team_status) == 1
                    assert match.player_tables[1].team_status[0].name == (
                        "Shrine of Maya"
                    )
                    assert match.player_tables[1].team_status[0].usage == 2
                    assert (
                        match.player_tables[1].characters[2].status[0].name
                        == "Seed of Skandha"
                    )
                    assert match.player_tables[1].characters[2].status[0].usage == 2
                elif test_id == 5:
                    characters = match.player_tables[0].characters
                    for cid, character in enumerate(characters):
                        assert len(character.status) == 1
                        assert character.status[0].name == "Seed of Skandha"
                        assert character.status[0].usage == 2
                        assert character.element_application == ["DENDRO"]
                    check_hp(match, [[3, 4, 5], [10, 10, 8]])
                elif test_id == 7:
                    # hp 4 1 1, ele DDN, all 1seed, burining flame, 1 maya
                    characters = match.player_tables[0].characters
                    for cid, character in enumerate(characters):
                        assert len(character.status) == 1
                        assert character.status[0].name == "Seed of Skandha"
                        assert character.status[0].usage == 1
                        assert character.element_application == ["DENDRO"]
                    check_hp(match, [[2, 3, 1], [10, 10, 8]])
                    assert len(match.player_tables[1].team_status) == 1
                    assert match.player_tables[1].team_status[0].name == (
                        "Shrine of Maya"
                    )
                    assert match.player_tables[1].team_status[0].usage == 1
                    assert len(match.player_tables[1].summons) == 1
                    assert match.player_tables[1].summons[0].name == ("Burning Flame")
                    assert match.player_tables[1].summons[0].usage == 1
                else:
                    break
            make_respond(agent_1, match)
        else:
            raise AssertionError("No need respond.")
        if len(agent_1.commands) == 0:
            break

    assert len(agent_1.commands) == 0
    assert match.round_number == 3
    check_hp(match, [[2, 3, 1], [10, 10, 8]])

    assert match.state != MatchState.ERROR


def test_nahida_apply_seed_to_defeated():
    """
    try to apply seed to defeated character.
    """
    agent_0 = NothingAgent(player_idx=0)
    agent_1 = InteractionAgent(
        player_idx=1,
        verbose_level=0,
        commands=[
            "sw_card",
            "choose 2",
            "skill 1 0 1 2",
            "skill 1 0 1 2",
            "skill 3 0 1 2",
            "skill 1 0 1 2",
            "end",
            "skill 1 0 1 2",
            "skill 2 0 1 2 3 4",
            "end",
        ],
        only_use_command=True,
    )
    match = Match(random_state=get_random_state())
    deck = Deck.from_str(
        """
        character:Fischl
        character:Mona
        character:Nahida
        Rana*10
        Wine-Stained Tricorne*10
        The Seed of Stored Knowledge*10
        """
    )
    match.set_deck([deck, deck])
    match.config.max_same_card_number = 30
    set_16_omni(match)
    assert match.start()[0]
    match.step()

    while True:
        if match.need_respond(0):
            make_respond(agent_0, match)
        elif match.need_respond(1):
            make_respond(agent_1, match)
        else:
            raise AssertionError("No need respond.")
        if len(agent_1.commands) == 0:
            break

    assert len(agent_1.commands) == 0
    assert match.round_number == 3
    check_hp(match, [[0, 5, 10], [10, 10, 10]])
    assert len(match.player_tables[1].team_status) == 0
    for character in match.player_tables[0].characters:
        if character.is_alive:
            assert len(character.status) == 1
            assert character.status[0].name == "Seed of Skandha"
            assert character.status[0].usage == 2
        else:
            assert len(character.status) == 0

    assert match.state != MatchState.ERROR


def test_maya_not_first_status():
    """
    AA + electro A + talent
    """
    agent_0 = NothingAgent(player_idx=0)
    agent_1 = InteractionAgent(
        player_idx=1,
        verbose_level=0,
        commands=[
            "sw_card",
            "choose 2",
            "skill 0 0 1 2",
            "skill 0 0 1 2",
            "sw_char 0 0",
            "skill 0 0 1 2",
            "sw_char 2 0",
            "end",
            "card 0 0 0 1 2",
            "TEST 1 maya 3",
            "end",
        ],
        only_use_command=True,
    )
    match = Match(random_state=get_random_state())
    deck = Deck.from_str(
        """
        character:ElectroMobMage
        character:Mona
        character:Nahida
        The Seed of Stored Knowledge*30
        """
    )
    match.set_deck([deck, deck])
    match.config.max_same_card_number = 30
    match.config.random_first_player = False
    set_16_omni(match)
    assert match.start()[0]
    match.step()

    while True:
        if match.need_respond(0):
            make_respond(agent_0, match)
        elif match.need_respond(1):
            while True:
                test_id = get_test_id_from_command(agent_1)
                if test_id == 1:
                    status = match.player_tables[1].team_status
                    found = False
                    for s in status:
                        if s.name == "Shrine of Maya":
                            assert s.usage == 3
                            found = True
                    assert found
                else:
                    break
            make_respond(agent_1, match)
        else:
            raise AssertionError("No need respond.")
        if len(agent_1.commands) == 0:
            break

    assert len(agent_1.commands) == 0
    assert match.round_number == 3
    check_hp(match, [[1, 10, 10], [10, 10, 10]])
    assert len(match.player_tables[1].team_status) == 2
    assert match.player_tables[1].team_status[1].name == "Shrine of Maya"
    assert match.player_tables[1].team_status[1].usage == 2
    assert match.player_tables[1].team_status[0].name == "Catalyzing Field"
    assert match.player_tables[1].team_status[0].usage == 1

    assert match.state != MatchState.ERROR


def test_seed_not_first_status():
    """
    AA + electro A + talent
    """
    agent_0 = NothingAgent(player_idx=0)
    agent_1 = InteractionAgent(
        player_idx=1,
        verbose_level=0,
        commands=[
            "sw_card",
            "choose 1",
            "skill 0 0 1 2",
            "sw_char 0 0",
            "skill 0 0 1 2",
            "sw_char 2 0",
            "skill 1 0 1 2",
            "skill 1 0 1 2",
            "sw_char 1 0",
            "skill 0 0 1 2",
            "sw_char 2 0",
        ],
        only_use_command=True,
    )
    match = Match(random_state=get_random_state())
    deck = Deck.from_str(
        """
        character:CryoMobMage
        character:Mona
        character:Nahida
        The Seed of Stored Knowledge*30
        """
    )
    match.set_deck([deck, deck])
    match.config.max_same_card_number = 30
    match.config.random_first_player = False
    set_16_omni(match)
    match.config.initial_dice_number = 32
    match.config.max_dice_number = 32
    assert match.start()[0]
    match.step()

    while True:
        if match.need_respond(0):
            make_respond(agent_0, match)
        elif match.need_respond(1):
            make_respond(agent_1, match)
        else:
            raise AssertionError("No need respond.")
        if len(agent_1.commands) == 0:
            break

    assert len(agent_1.commands) == 0
    assert match.round_number == 1
    check_hp(match, [[0, 9, 9], [10, 10, 10]])
    characters = match.player_tables[0].characters
    assert len(characters[0].status) == 0
    assert len(characters[1].status) == 1
    assert characters[1].status[0].name == "Seed of Skandha"
    assert characters[1].status[0].usage == 1
    assert len(characters[2].status) == 1
    assert characters[2].status[0].name == "Seed of Skandha"
    assert characters[2].status[0].usage == 1

    assert match.state != MatchState.ERROR


def test_talent_enemy_has_other_character_status():
    """
    when enemy has other character status, equipping talent should only
    add usage of seed.
    in round 3, player 0 will do elemental reaction but will not increase
    damage.
    """
    agent_0 = InteractionAgent(
        player_idx=0,
        verbose_level=0,
        commands=[
            "sw_card",
            "choose 0",
            "end",
            "end",
            "skill 0 0 1 2",
            "sw_char 1 0",
            "skill 0 0 1 2",
        ],
        only_use_command=True,
    )
    agent_1 = InteractionAgent(
        player_idx=1,
        verbose_level=0,
        commands=[
            "sw_card",
            "choose 1",
            "skill 0 0 1 2",
            "sw_char 3 0",
            "skill 1 0 1 2",
            "skill 1 0 1 2",
            "sw_char 2 0",
            "end",
            "skill 0 0 1 2",
            "sw_char 3 0",
            "card 0 0 0 1 2",
            "TEST 1 maya 3 seed all 2 frozen 1",
            "skill 1 0 1 2",
            "end",
            "end",
        ],
        only_use_command=True,
    )
    match = Match(random_state=get_random_state())
    deck = Deck.from_str(
        """
        character:ElectroMobMage
        character:CryoMobMage
        character:Mona
        character:Nahida
        The Seed of Stored Knowledge*30
        """
    )
    # add HP to avoid defeated.
    deck.characters[0].hp = 90
    deck.characters[0].max_hp = 90
    match.set_deck([deck, deck])
    match.config.max_same_card_number = 30
    match.config.character_number = 4
    match.config.random_first_player = False
    set_16_omni(match)
    # match.enable_history = True
    assert match.start()[0]
    match.step()

    while True:
        if match.need_respond(0):
            make_respond(agent_0, match)
        elif match.need_respond(1):
            while True:
                test_id = get_test_id_from_command(agent_1)
                if test_id == 1:
                    status = match.player_tables[1].team_status
                    found = False
                    for s in status:
                        if s.name == "Shrine of Maya":  # pragma: no branch
                            assert s.usage == 3
                            found = True
                    assert found
                    characters = match.player_tables[0].characters
                    for character in characters:
                        assert character.status[0].name == "Seed of Skandha"
                        assert character.status[0].usage == 2
                    assert len(characters[0].status) == 2
                    assert characters[0].status[1].name == "Frozen"
                    assert characters[0].status[1].usage == 1
                else:
                    break
            make_respond(agent_1, match)
        else:
            raise AssertionError("No need respond.")
        if len(agent_1.commands) == 0 and len(agent_0.commands) == 0:
            break

    assert len(agent_1.commands) == 0 and len(agent_0.commands) == 0
    assert match.round_number == 3
    check_hp(match, [[76, 9, 9, 9], [89, 9, 9, 7]])
    assert len(match.player_tables[1].team_status) == 1
    assert match.player_tables[1].team_status[0].name == "Shrine of Maya"
    assert match.player_tables[1].team_status[0].usage == 2

    assert match.state != MatchState.ERROR


def test_seed_with_element_application():
    cmd_records = [
        [
            "sw_card",
            "choose 0",
            "sw_char 2 0",
            "sw_char 1 0",
            "skill 2 0 1 2 3 4",
            "TEST 2 p0 team usage 1",
            "TEST 3 p1 character usage 2",
            "TEST 1 10 6 10 7 10 10",
            "end",
        ],
        ["sw_card", "choose 1", "sw_char 0 0", "skill 0 0 1 2", "skill 1 0 1 2"],
    ]
    agent_0 = InteractionAgent(
        player_idx=0, verbose_level=0, commands=cmd_records[0], only_use_command=True
    )
    agent_1 = InteractionAgent(
        player_idx=1, verbose_level=0, commands=cmd_records[1], only_use_command=True
    )
    # initialize match. It is recommended to use default random state to make
    # replay unchanged.
    match = Match(random_state=get_random_state())
    # deck information
    deck = Deck.from_str(
        """
        default_version:4.0
        character:Xingqiu@3.3
        character:Nahida
        character:Fischl
        The Scent Remained*30
        """
    )
    match.set_deck([deck, deck])
    match.config.max_same_card_number = None
    match.config.character_number = None
    match.config.card_number = None
    match.config.check_deck_restriction = False
    # check whether random_first_player is enabled.
    match.config.random_first_player = False
    # check whether in rich mode (16 omni each round)
    set_16_omni(match)
    match.start()
    match.step()

    while True:
        if match.need_respond(0):
            agent = agent_0
        elif match.need_respond(1):
            agent = agent_1
        else:
            raise AssertionError("No need respond.")
        # do tests
        while True:
            cmd = agent.commands[0]
            test_id = get_test_id_from_command(agent)
            if test_id == 0:
                # id 0 means current command is not a test command.
                break
            elif test_id == 1:
                # a sample of HP check based on the command string.
                hps = cmd.strip().split(" ")[2:]
                hps = [int(x) for x in hps]
                hps = [hps[:3], hps[3:]]
                check_hp(match, hps)
            elif test_id == 2:
                assert len(match.player_tables[0].team_status) == 1
                assert match.player_tables[0].team_status[0].usage == 1
            elif test_id == 3:
                for c in match.player_tables[1].characters:
                    assert len(c.status) == 1
                    assert c.status[0].usage == 2
            else:
                raise AssertionError(f"Unknown test id {test_id}")
        # respond
        make_respond(agent, match)
        if len(agent_1.commands) == 0 and len(agent_0.commands) == 0:
            break

    # simulate ends, check final state
    assert match.state != MatchState.ERROR


def test_nahida_defeat_with_last_seed():
    cmd_records = [
        ["sw_card", "choose 2", "end", "end", "choose 0"],
        [
            "sw_card",
            "choose 1",
            "skill 1 15 14 13",
            "skill 1 12 11 10",
            "sw_char 2 9",
            "skill 1 8 7 6",
            "sw_char 1 5",
            "end",
            "skill 3 15 14 13",
            "TEST 1 8 8 0 10 10 10",
            "end",
        ],
    ]
    agent_0 = InteractionAgent(
        player_idx=0, verbose_level=0, commands=cmd_records[0], only_use_command=True
    )
    agent_1 = InteractionAgent(
        player_idx=1, verbose_level=0, commands=cmd_records[1], only_use_command=True
    )
    # initialize match. It is recommended to use default random state to make
    # replay unchanged.
    match = Match(random_state=get_random_state())
    # deck information
    deck = Deck.from_str(
        """
        default_version:4.0
        character:Xingqiu@3.3
        character:Nahida
        character:Fischl
        The Scent Remained*30
        """
    )
    match.set_deck([deck, deck])
    match.config.max_same_card_number = None
    match.config.character_number = None
    match.config.card_number = None
    match.config.check_deck_restriction = False
    # check whether random_first_player is enabled.
    match.config.random_first_player = False
    # check whether in rich mode (16 omni each round)
    set_16_omni(match)
    match.start()
    match.step()

    while True:
        if match.need_respond(0):
            agent = agent_0
        elif match.need_respond(1):
            agent = agent_1
        else:
            raise AssertionError("No need respond.")
        # do tests
        while True:
            cmd = agent.commands[0]
            test_id = get_test_id_from_command(agent)
            if test_id == 0:
                # id 0 means current command is not a test command.
                break
            elif test_id == 1:
                # a sample of HP check based on the command string.
                hps = cmd.strip().split(" ")[2:]
                hps = [int(x) for x in hps]
                hps = [hps[:3], hps[3:]]
                check_hp(match, hps)
            else:
                raise AssertionError(f"Unknown test id {test_id}")
        # respond
        make_respond(agent, match)
        if len(agent_1.commands) == 0 and len(agent_0.commands) == 0:
            break

    # simulate ends, check final state
    assert match.state != MatchState.ERROR


if __name__ == "__main__":
    # test_fischl_mona_nahida()
    # test_seed_not_first_status()
    # test_talent_enemy_has_other_character_status()
    # test_nahida_talents()
    # test_seed_with_element_application()
    # test_fischl_mona_nahida_no_talent()
    test_nahida_defeat_with_last_seed()



from typing import Any, List, Literal, Tuple

from pydantic import PrivateAttr

from ....utils.class_registry import register_class

from ...summon.base import DefendSummonBase, AttackerSummonBase

from ...event import SkillEndEventArguments

from ...action import (
    Actions,
    ChangeObjectUsageAction,
    ChargeAction,
    CreateRandomObjectAction,
)
from ...struct import Cost, ObjectPosition

from ...consts import (
    ELEMENT_TO_DAMAGE_TYPE,
    DamageElementalType,
    DieColor,
    ElementType,
    FactionType,
    ObjectPositionType,
    SkillType,
    WeaponType,
)
from ..character_base import (
    ElementalBurstBase,
    ElementalNormalAttackBase,
    ElementalSkillBase,
    CharacterBase,
    SkillBase,
    SkillTalent,
)


# Summons


class Squirrel_3_3(AttackerSummonBase):
    name: Literal["Oceanic Mimic: Squirrel"] = "Oceanic Mimic: Squirrel"
    version: Literal["3.3"] = "3.3"
    usage: int = 2
    max_usage: int = 2
    damage_elemental_type: DamageElementalType = DamageElementalType.HYDRO
    damage: int = 2


class Raptor_3_3(AttackerSummonBase):
    name: Literal["Oceanic Mimic: Raptor"] = "Oceanic Mimic: Raptor"
    version: Literal["3.3"] = "3.3"
    usage: int = 3
    max_usage: int = 3
    damage_elemental_type: DamageElementalType = DamageElementalType.HYDRO
    damage: int = 1


class Frog_3_3(DefendSummonBase):
    name: Literal["Oceanic Mimic: Frog"] = "Oceanic Mimic: Frog"
    version: Literal["3.3"] = "3.3"
    usage: int = 2
    max_usage: int = 2
    damage_elemental_type: DamageElementalType = DamageElementalType.HYDRO
    damage: int = 2
    min_damage_to_trigger: int = 1
    max_in_one_time: int = 1
    attack_until_run_out_of_usage: bool = True


# Skills


mimic_names = [
    "Oceanic Mimic: Squirrel",
    "Oceanic Mimic: Frog",
    "Oceanic Mimic: Raptor",
]


class RhodeiaElementSkill(ElementalSkillBase):
    name: Literal["Oceanid Mimic Summoning", "The Myriad Wilds"]
    damage: int = 0
    damage_type: DamageElementalType = DamageElementalType.HYDRO
    cost: Cost = Cost(elemental_dice_color=DieColor.HYDRO, elemental_dice_number=3)
    _summon_number: int = PrivateAttr(1)
    version: Literal["3.3"] = "3.3"

    def __init__(self, *argv, **kwargs):
        super().__init__(*argv, **kwargs)
        if self.name == "Oceanid Mimic Summoning":
            pass
        else:
            assert self.name == "The Myriad Wilds"
            self._summon_number = 2
            self.cost.elemental_dice_number = 5

    def _get_exist_unexist_names(self, match: Any) -> Tuple[List[str], List[str]]:
        """
        Create two lists including exist and unexist names.
        """
        summons = match.player_tables[self.position.player_idx].summons
        summon_names = [x.name for x in summons]
        unexist_names: List[str] = []
        exist_names: List[str] = []
        for name in mimic_names:
            if name not in summon_names:
                unexist_names.append(name)
            else:
                exist_names.append(name)
        return exist_names, unexist_names

    def get_actions(self, match: Any) -> List[ChargeAction | CreateRandomObjectAction]:
        """
        create object
        """
        exist_names, unexist_names = self._get_exist_unexist_names(match)
        ret: List[ChargeAction | CreateRandomObjectAction] = []
        target_position = ObjectPosition(
            player_idx=self.position.player_idx, area=ObjectPositionType.SUMMON, id=-1
        )
        if len(unexist_names) == 0:
            # all exist, only from exist
            ret.append(
                CreateRandomObjectAction(
                    object_names=exist_names,
                    object_position=target_position,
                    object_arguments={"version": self.version},
                    number=self._summon_number,
                )
            )
        elif len(unexist_names) >= self._summon_number:
            # enough unexist
            ret.append(
                CreateRandomObjectAction(
                    object_names=unexist_names,
                    object_position=target_position,
                    object_arguments={"version": self.version},
                    number=self._summon_number,
                )
            )
        else:
            # not enough unexist, create all unexist and refresh random exist
            ret.append(
                CreateRandomObjectAction(
                    object_names=unexist_names,
                    object_position=target_position,
                    object_arguments={"version": self.version},
                    number=len(unexist_names),
                )
            )
            ret.append(
                CreateRandomObjectAction(
                    object_names=exist_names,
                    object_position=target_position,
                    object_arguments={"version": self.version},
                    number=self._summon_number - len(unexist_names),
                )
            )
        ret.append(self.charge_self(1))
        return ret


class TideAndTorrent(ElementalBurstBase):
    name: Literal["Tide and Torrent"] = "Tide and Torrent"
    damage: int = 2
    damage_per_summon: int = 2
    damage_type: DamageElementalType = DamageElementalType.HYDRO
    cost: Cost = Cost(
        elemental_dice_color=DieColor.HYDRO, elemental_dice_number=3, charge=3
    )

    def get_actions(self, match: Any) -> List[Actions]:
        summons = match.player_tables[self.position.player_idx].summons
        self.damage += len(summons) * self.damage_per_summon
        ret = super().get_actions(match)
        self.damage -= len(summons) * self.damage_per_summon
        return ret


# Talents


class StreamingSurge_3_3(SkillTalent):
    name: Literal["Streaming Surge"]
    version: Literal["3.3"] = "3.3"
    character_name: Literal["Rhodeia of Loch"] = "Rhodeia of Loch"
    cost: Cost = Cost(
        elemental_dice_color=DieColor.HYDRO, elemental_dice_number=4, charge=3
    )
    skill: Literal["Tide and Torrent"] = "Tide and Torrent"

    def event_handler_SKILL_END(
        self, event: SkillEndEventArguments, match: Any
    ) -> List[ChangeObjectUsageAction]:
        """
        When equipped and this character use burst, change all summons' usage
        by 1.
        """
        if not self.position.check_position_valid(
            event.action.position,
            match,
            player_idx_same=True,
            character_idx_same=True,
            source_area=ObjectPositionType.CHARACTER,
            target_area=ObjectPositionType.SKILL,
        ):
            # not equipped, or not self use skill, or not skill
            return []
        skill: SkillBase = match.get_object(event.action.position)
        if skill.skill_type != SkillType.ELEMENTAL_BURST:
            # not elemental burst
            return []
        summons = match.player_tables[self.position.player_idx].summons
        ret = []
        for summon in summons:
            ret.append(
                ChangeObjectUsageAction(object_position=summon.position, change_usage=1)
            )
        return ret


# character base


class RhodeiaOfLoch_3_3(CharacterBase):
    name: Literal["Rhodeia of Loch"]
    version: Literal["3.3"] = "3.3"
    element: ElementType = ElementType.HYDRO
    max_hp: int = 10
    max_charge: int = 3
    skills: List[ElementalNormalAttackBase | RhodeiaElementSkill | TideAndTorrent] = []
    faction: List[FactionType] = [FactionType.MONSTER]
    weapon_type: WeaponType = WeaponType.OTHER

    def _init_skills(self) -> None:
        self.skills = [
            ElementalNormalAttackBase(
                name="Surge",
                damage_type=ELEMENT_TO_DAMAGE_TYPE[self.element],
                cost=ElementalNormalAttackBase.get_cost(self.element),
            ),
            RhodeiaElementSkill(
                name="Oceanid Mimic Summoning",
            ),
            RhodeiaElementSkill(
                name="The Myriad Wilds",
            ),
            TideAndTorrent(),
        ]


register_class(
    RhodeiaOfLoch_3_3 | Squirrel_3_3 | Raptor_3_3 | Frog_3_3 | StreamingSurge_3_3
)



from typing import Dict, Literal, List

from lpsim.server.consts import ELEMENT_TO_DIE_COLOR
from lpsim.server.action import CreateDiceAction
from lpsim.server.patch.v43.cards.gilded_dreams_4_3 import GildedDreams_4_3
from lpsim.server.struct import Cost
from lpsim.server import Match
from lpsim.utils.class_registry import register_class
from lpsim.utils.desc_registry import DescDictType


class GildedDreams_4_5(GildedDreams_4_3):
    version: Literal["4.5"] = "4.5"
    cost: Cost = Cost(same_dice_number=3)

    def equip(self, match: Match) -> List[CreateDiceAction]:
        """
        When equipped, create one elemental dice. If there are 3 different
        elemental dice in your team, create 2 dice.
        """
        super().equip(match)
        characters = match.player_tables[self.position.player_idx].characters
        equip_character = characters[self.position.character_idx]
        ret: CreateDiceAction = CreateDiceAction(
            player_idx=self.position.player_idx,
            number=1,
            color=ELEMENT_TO_DIE_COLOR[equip_character.element],
        )
        elements = set()
        for character in characters:
            elements.add(character.element)
        if len(elements) >= 3:
            ret.number = 2
        return [ret]


desc: Dict[str, DescDictType] = {
    "ARTIFACT/Gilded Dreams": {
        "descs": {
            "4.5": {
                "en-US": "When played: generate 1 Die of the same Element as the attached character. If you have 3 different Elemental Types in your party, generate 2 Dice of the same Element as the attached character.\nWhen the character to which this card is attached is your active character, then when an opposing character takes Elemental Reaction DMG: Draw a card. (Twice per Round)\n(A character can equip a maximum of 1 Artifact)",  # noqa: E501
                "zh-CN": "入场时：生成1个所附属角色类型的元素骰。如果我方队伍中存在3种不同元素类型的角色，则生成2个所附属角色类型的元素骰。\n所附属角色为出战角色期间，敌方受到元素反应伤害时：抓1张牌。（每回合至多2次）\n（角色最多装备1件「圣遗物」）",  # noqa: E501
            }
        },
    },
}


register_class(GildedDreams_4_5, desc)
